import logging

from functools import partial
from gcloud_requests import DatastoreRequestsProxy, enter_transaction, exit_transaction
from google.cloud import datastore
from threading import local

from .. import Adapter, Key
from ..adapter import QueryResponse
from ..model import KeyLike
from ..transaction import Transaction, TransactionFailed

_logger = logging.getLogger(__name__)


class _DeferredKey(KeyLike):
    def __init__(self, ds_entity):
        self.ds_entity = ds_entity
        self._value = None

    @property
    def _anom_key(self):
        if self._value is None or self._value.is_partial:
            self._value = DatastoreAdapter._convert_key_from_datastore(self.ds_entity.key)
        return self._value

    def __getattr__(self, name):
        return getattr(self._anom_key, name)

    def __repr__(self):
        return repr(self._anom_key)


class _DatastoreOuterTransaction(Transaction):
    def __init__(self, adapter):
        self.adapter = adapter
        self.ds_transaction = adapter.client.transaction()

    def begin(self):
        _logger.debug("Beginning transaction...")
        self.ds_transaction.begin()
        self.adapter.client._push_batch(self.ds_transaction)
        enter_transaction()

    def commit(self):
        try:
            _logger.debug("Committing transaction...")
            self.ds_transaction.commit()
        except Exception as e:
            _logger.debug("Transaction failed: %s", e)
            raise TransactionFailed("Failed to commit transaction.", cause=e)

    def rollback(self):
        _logger.debug("Rolling transaction back...")
        self.ds_transaction.rollback()

    def end(self):
        _logger.debug("Ending transaction...")
        exit_transaction()
        self.adapter.client._pop_batch()
        self.adapter._transactions.remove(self)


class _DatastoreInnerTransaction(Transaction):
    def __init__(self, parent):
        self.parent = parent

    def begin(self):
        _logger.debug("Beginning inner transaction...")

    def commit(self):
        _logger.debug("Committing inner transaction...")

    def rollback(self):
        _logger.debug("Rolling back inner transaction...")

    def end(self):
        _logger.debug("Ending inner transaction...")
        self.adapter._transactions.remove(self)

    def __getattr__(self, name):
        return getattr(self.parent, name)


class DatastoreAdapter(Adapter):
    """A Google Cloud Datastore adapter based on :mod:`google.cloud.datastore`.

    Parameters:
      project(str, optional): The project this Adapter should connect to.
        Adapter should operate by default.  Individual Datastore Keys
        may specify their own namespaces and override this.
      credentials(datastore.Credentials): The OAuth2 Credentials to
        use for this client.  If not passed, falls back to the default
        inferred from the environment.
    """

    _state = local()

    def __init__(self, *, project=None, credentials=None):
        self.project = project
        self.credentials = credentials
        self.proxy = DatastoreRequestsProxy(credentials=credentials)
        self.client = datastore.Client(
            credentials=self.credentials,
            project=self.project,
            _http=self.proxy,
            _use_grpc=False,
        )

    @property
    def _transactions(self):
        "list[Transaction]: The current stack of Transactions."
        transactions = getattr(self._state, "transactions", None)
        if transactions is None:
            transactions = self._state.transactions = []
        return transactions

    def delete_multi(self, keys):
        self.client.delete_multi(self._convert_key_to_datastore(key) for key in keys)

    def get_multi(self, keys):
        get_multi = self.client.get_multi
        if self.in_transaction:
            transaction = self.current_transaction
            get_multi = partial(get_multi, transaction=transaction.ds_transaction)

        datastore_keys = [self._convert_key_to_datastore(key) for key in keys]
        request_keys = set(datastore_keys)
        found, missing, deferred = [], [], []
        while True:
            found.extend(get_multi(request_keys, missing=missing, deferred=deferred))
            if not deferred:
                break

            for entity in found:  # pragma: no cover
                request_keys.remove(entity.key)

            for key in missing:  # pragma: no cover
                request_keys.remove(key)

        results = [None] * len(keys)
        for entity in found:
            index = datastore_keys.index(entity.key)
            results[index] = self._prepare_to_load(entity)

        return results

    def put_multi(self, requests):
        entities = [self._prepare_to_store(*request) for request in requests]
        self.client.put_multi(entities)
        if self.in_transaction:
            return [_DeferredKey(entity) for entity in entities]
        return [self._convert_key_from_datastore(entity.key) for entity in entities]

    def query(self, query, options):
        ancestor = None
        if query.ancestor:
            ancestor = self._convert_key_to_datastore(query.ancestor)

        filters = self._convert_filters_to_datastore(query.filters)
        query = self.client.query(
            kind=query.kind,
            ancestor=ancestor,
            namespace=query.namespace,
            projection=query.projection,
            order=query.orders,
            filters=filters,
        )
        if options.keys_only:
            query.keys_only()

        result_iterator = query.fetch(
            limit=options.batch_size,
            offset=options.offset,
            start_cursor=options.cursor,
        )

        entities = []
        for entity in result_iterator:
            key, data = self._convert_key_from_datastore(entity.key), None
            if not options.keys_only:
                data = self._prepare_to_load(entity)

            entities.append((key, data))

        return QueryResponse(entities=entities, cursor=result_iterator.next_page_token)

    def transaction(self, propagation):
        if propagation == Transaction.Propagation.Independent:
            transaction = _DatastoreOuterTransaction(self)
            self._transactions.append(transaction)
            return transaction

        elif propagation == Transaction.Propagation.Nested:
            if self._transactions:
                transaction = _DatastoreInnerTransaction(self.current_transaction)
            else:
                transaction = _DatastoreOuterTransaction(self)

            self._transactions.append(transaction)
            return transaction

        else:  # pragma: no cover
            raise ValueError(f"Invalid propagation option {propagation!r}.")

    @property
    def in_transaction(self):
        return bool(self._transactions)

    @property
    def current_transaction(self):
        return self._transactions[-1]

    def _convert_filters_to_datastore(self, filters):
        for property_filter in filters:
            prop, op, value = property_filter
            if isinstance(value, Key):
                value = self._convert_key_to_datastore(property_filter.value)

            yield prop, op, value

    def _convert_key_to_datastore(self, anom_key):
        return self.client.key(*anom_key.path, namespace=anom_key.namespace or None)

    @staticmethod
    def _convert_key_from_datastore(datastore_key):
        return Key.from_path(*datastore_key.flat_path, namespace=datastore_key.namespace)

    def _prepare_to_store(self, key, unindexed, data):
        datastore_key = self._convert_key_to_datastore(key)
        entity = datastore.Entity(datastore_key, unindexed)
        entity.update({name: self._prepare_to_store_value(value) for name, value in data})
        return entity

    def _prepare_to_store_value(self, value):
        if isinstance(value, Key):
            return self._convert_key_to_datastore(value)

        elif isinstance(value, (tuple, list)):
            return [self._prepare_to_store_value(v) for v in value]

        else:
            return value

    def _prepare_to_load(self, entity):
        data = {}
        for name, value in entity.items():
            if isinstance(value, datastore.Key):
                value = self._convert_key_from_datastore(value)

            elif isinstance(value, list) and all(isinstance(v, datastore.Key) for v in value):
                value = [self._convert_key_from_datastore(v) for v in value]

            data[name] = value

        return data



# flake8: noqa
from . import conditions, properties, properties as props
from .adapter import Adapter, get_adapter, set_adapter
from .model import Key, Model, Property, delete_multi, get_multi, put_multi, lookup_model_by_kind
from .namespaces import get_namespace, namespace, set_default_namespace, set_namespace
from .query import Query, Resultset, Page, Pages
from .transaction import Transaction, TransactionError, RetriesExceeded, transactional

__version__ = "0.9.1"



from collections import namedtuple
from threading import RLock
from weakref import WeakValueDictionary

from .adapter import PutRequest, get_adapter
from .namespaces import get_namespace
from .query import PropertyFilter, Query

#: The set of known models.  This is used to look up model classes at
#: runtime by their kind.
_known_models = WeakValueDictionary()
_known_models_lock = RLock()

#: Canary value for unset properties.
NotFound = type("NotFound", (object,), {})()

#: Canary value for properties that should be skipped when loading
#: entities from Datastore.
Skip = type("Skip", (object,), {})()


def classname(ob):
    "Returns the name of ob's class."
    return type(ob).__name__


class KeyLike:
    """Base class for objects that should be treated as if they are
    datastore keys (for example, when comparing two objects with one
    another).
    """


class Key(KeyLike, namedtuple("Key", ("kind", "id_or_name", "parent", "namespace"))):
    """A Datastore key.

    Parameters:
      kind(str or model): The Datastore kind this key represents.
      id_or_name(int or str): The id or name of this key.
      parent(anom.Key, optional): This key's ancestor.
      namespace(str, optional): This key's namespace.

    Attributes:
      kind(str): This key's kind.
      id_or_name(int or str or None): This key's integer id or string
        name.  This is ``None`` for partial keys.
      parent(anom.Key or None): This key's ancestor.
      namespace(str or None): This key's namespace.
    """

    def __new__(cls, kind, id_or_name=None, parent=None, namespace=None):
        if isinstance(kind, model):
            kind = kind._kind

        if parent and parent.is_partial:
            raise ValueError("Cannot use partial Keys as parents.")

        if parent:
            if namespace is not None and namespace != parent.namespace:
                raise ValueError(f"Namespace {namespace!r} is different from parent namespace {parent.namespace!r}.")

            namespace = parent.namespace

        elif namespace is None:
            namespace = get_namespace()

        return super().__new__(cls, kind, id_or_name, parent, namespace)

    @classmethod
    def from_path(cls, *path, namespace=None):
        """Build up a Datastore key from a path.

        Parameters:
          \*path(tuple[str or int]): The path segments.
          namespace(str): An optional namespace for the key. This is
            applied to each key in the tree.

        Returns:
          anom.Key: The Datastore represented by the given path.
        """
        parent = None
        for i in range(0, len(path), 2):
            parent = cls(*path[i:i + 2], parent=parent, namespace=namespace)

        return parent

    @property
    def path(self):
        "tuple: The full Datastore path represented by this key."
        prefix = ()
        if self.parent:
            prefix = self.parent.path

        if self.id_or_name:
            return prefix + (self.kind, self.id_or_name)

        return prefix + (self.kind,)

    @property
    def is_partial(self):
        "bool: ``True`` if this key doesn't have an id yet."
        return len(self.path) % 2 != 0

    @property
    def int_id(self):
        "int: This key's numeric id."
        id_or_name = self.id_or_name
        if id_or_name is not None and isinstance(id_or_name, int):
            return id_or_name
        return None

    @property
    def str_id(self):
        "str: This key's string id."
        id_or_name = self.id_or_name
        if id_or_name is not None and isinstance(id_or_name, str):
            return id_or_name
        return None

    def get_model(self):
        """Get the model class for this Key.

        Raises:
          RuntimeError: If a model isn't registered for the Key's
            kind.

        Returns:
          type: A Model class.
        """
        return lookup_model_by_kind(self.kind)

    def delete(self):
        """Delete the entity represented by this Key from Datastore.
        """
        return delete_multi([self])

    def get(self):
        """Get the entity represented by this Key from Datastore.

        Returns:
          Model: The entity or ``None`` if it does not exist.
        """
        return get_multi([self])[0]

    def __repr__(self):
        return f"Key({self.kind!r}, {self.id_or_name!r}, parent={self.parent!r}, namespace={self.namespace!r})"

    def __hash__(self):
        return hash(tuple(self))

    def __eq__(self, other):
        if not isinstance(other, KeyLike):
            return False

        if self is other:
            return True

        if other.parent != self.parent or \
           other.namespace != self.namespace:
            return False

        return self.path == other.path

    def __ne__(self, other):
        return not (self == other)


class Property:
    """Base class for Datastore model properties.

    The property lifecycle is as follows:

    * :meth:`__init__` is called every time a property is defined on a
      model class.
    * :meth:`validate` is called every time a value is assigned to a
      property on a model instance.
    * :meth:`prepare_to_load` is called before a property is assigned
      to a model instance that is being loaded from Datastore.
    * :meth:`prepare_to_store` is called before a property is
      persisted from a model instance to Datastore.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required-but-empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    #: The types of values that may be assigned to this property.
    _types = ()

    def __init__(self, *, name=None, default=None, indexed=False, indexed_if=None, optional=False, repeated=False):
        self.indexed = indexed or bool(indexed_if)
        self.indexed_if = indexed_if
        self.optional = optional
        self.repeated = repeated

        self.default = self.validate(default) if default is not None else None

        self._name_on_entity = name
        self._name_on_model = None

    @property
    def name_on_entity(self):
        "str: The name of this Property inside the Datastore entity."
        return self._name_on_entity

    @property
    def name_on_model(self):
        "str: The name of this Property on the Model instance."
        return self._name_on_model

    @property
    def is_none(self):
        "PropertyFilter: A filter that checks if this value is None."
        if not self.optional:
            raise TypeError("Required properties cannot be compared against None.")
        return self == None  # noqa

    def validate(self, value):
        """Validates that `value` can be assigned to this Property.

        Parameters:
          value: The value to validate.

        Raises:
          TypeError: If the type of the assigned value is invalid.

        Returns:
          The value that should be assigned to the entity.
        """
        if isinstance(value, self._types):
            return value

        elif self.optional and value is None:
            return [] if self.repeated else None

        elif self.repeated and isinstance(value, (tuple, list)) and all(isinstance(x, self._types) for x in value):
            return value

        else:
            raise TypeError(f"Value of type {classname(value)} assigned to {classname(self)} property.")

    def prepare_to_load(self, entity, value):
        """Prepare `value` to be loaded into a Model.  Called by the
        model for each Property, value pair contained in the persisted
        data when loading it from an adapter.

        Parameters:
          entity(Model): The entity to which the value belongs.
          value: The value being loaded.

        Returns:
          The value that should be assigned to the entity.
        """
        return value

    def prepare_to_store(self, entity, value):
        """Prepare `value` for storage.  Called by the Model for each
        Property, value pair it contains before handing the data off
        to an adapter.

        Parameters:
          entity(Model): The entity to which the value belongs.
          value: The value being stored.

        Raises:
          RuntimeError: If this property is required but no value was
            assigned to it.

        Returns:
          The value that should be persisted.
        """
        if value is None and not self.optional:
            raise RuntimeError(f"Property {self.name_on_model} requires a value.")
        return value

    def __set_name__(self, ob, name):
        self._name_on_entity = self.name_on_entity or name
        self._name_on_model = name

    def __get__(self, ob, obtype):
        if ob is None:
            return self

        value = ob._data.get(self.name_on_model, NotFound)
        if value is NotFound:
            if self.default is not None:
                return self.default

            elif self.repeated:
                value = ob._data[self.name_on_model] = []

            else:
                return None

        return value

    def __set__(self, ob, value):
        ob._data[self.name_on_model] = self.validate(value)

    def __delete__(self, ob):
        del ob._data[self.name_on_model]

    def _build_filter(self, op, value):
        if not self.indexed:
            raise TypeError(f"{self.name_on_model} is not indexed.")
        return PropertyFilter(self.name_on_entity, op, self.validate(value))

    def __eq__(self, value):
        return self._build_filter("=", value)

    def __le__(self, value):
        return self._build_filter("<=", value)

    def __ge__(self, value):
        return self._build_filter(">=", value)

    def __lt__(self, value):
        return self._build_filter("<", value)

    def __gt__(self, value):
        return self._build_filter(">", value)

    def __neg__(self):
        return "-" + self.name_on_entity

    def __pos__(self):
        return self.name_on_entity


class EmbedLike(Property):
    """Base class for properties that Embed other models.
    """

    def get_unindexed_properties(self, entity):  # pragma: no cover
        """tuple[str]: The set of unindexed properties belonging to
        the underlying model and all nested models.
        """
        raise NotImplementedError


class _adapter:
    def __get__(self, ob, obtype):
        return get_adapter()


class model(type):
    """Metaclass of Model classes.

    Parameters:
      poly(bool, optional): Determines if the model should be
        polymorphic or not.  Subclasses of polymorphic models are all
        stored under the same kind.

    Attributes:
      _adapter(Adapter): A computed property that returns the adapter
        for this model class.
      _is_child(bool): Whether or not this is a child model in a
        polymorphic hierarchy.
      _is_root(bool): Whether or not this is the root model in a
        polymorphic hierarchy.
      _kind(str): The underlying Datastore kind of this model.
      _kinds(list[str]): The list of kinds in this model's hierarchy.
      _properties(dict): A dict of all of the properties defined on
        this model.
    """

    #: The name of the field that holds the flattened class hierarchy
    #: on polymodel entities.
    _kinds_name = "^k"

    def __new__(cls, classname, bases, attrs, poly=False, **kwargs):
        attrs["_adapter"] = _adapter()
        attrs["_is_child"] = is_child = False
        attrs["_is_root"] = poly
        attrs["_kind"] = kind = attrs.pop("_kind", classname)
        attrs["_kinds"] = kinds = [kind]

        # Collect all of the properties defined on this model.
        attrs["_properties"] = properties = {}
        for name, value in attrs.items():
            if isinstance(value, Property):
                properties[name] = value

        # Collect all of the properties and kinds defined on parents
        # of this model.
        for base in bases:
            if not isinstance(base, model):
                continue

            # Avoid adding the base Model class to kinds.
            if "Model" in globals() and base is Model:
                continue

            kinds.extend(base._kinds)
            if base._is_polymorphic:  # Poly bases "infect" everything below them.
                attrs["_is_child"] = is_child = True
                attrs["_kind"] = base._kind

            for name, prop in base._properties.items():
                if name not in properties:
                    properties[name] = prop

        clazz = type.__new__(cls, classname, bases, attrs)

        # Ensure that a single model maps to a single kind at runtime.
        with _known_models_lock:
            if kind in _known_models and not is_child:
                raise TypeError(f"Multiple models for kind {kind!r}.")

            _known_models[kind] = clazz

        return clazz

    @property
    def _is_polymorphic(self):
        "bool: True if this child belongs to a polymorphic hierarchy."
        return self._is_root or self._is_child


class Model(metaclass=model):
    """Base class for Datastore models.

    Attributes:
      key(anom.Key): The Datastore Key for this entity.  If the entity
        was never stored then the Key is going to be partial.

    Note:
      Hooks are only called when dealing with individual entities via
      their keys.  They *do not* run when entities are loaded from a
      query.
    """

    def __init__(self, *, key=None, **properties):
        self.key = key or Key(self._kind)

        self._data = {}
        for name, value in properties.items():
            if name not in self._properties:
                raise TypeError(f"{classname(self)}() does not take a {name!r} parameter.")

            setattr(self, name, value)

    def __iter__(self):
        for name, prop in self._properties.items():
            value = getattr(self, name)
            if isinstance(prop, EmbedLike):
                yield from prop.prepare_to_store(self, value)
            else:
                yield prop.name_on_entity, prop.prepare_to_store(self, value)

        # Polymorphic models need to keep track of their bases.
        if type(self)._is_polymorphic:
            yield model._kinds_name, self._kinds

    @classmethod
    def _load(cls, key, data):
        # Polymorphic models need to instantiate leaf classes.
        if cls._is_polymorphic and model._kinds_name in data:
            name = data[model._kinds_name][0]
            cls = lookup_model_by_kind(name)

        instance = cls()
        instance.key = key

        for name, prop in instance._properties.items():
            if isinstance(prop, EmbedLike):
                instance._data[name] = prop.prepare_to_load(instance, data)

            else:
                value = prop.prepare_to_load(instance, data.get(name))
                if value is not Skip:
                    instance._data[name] = value

        return instance

    @property
    def unindexed_properties(self):
        "tuple[str]: The names of all the unindexed properties on this entity."
        properties = ()
        for name, prop in self._properties.items():
            if isinstance(prop, EmbedLike):
                embedded_entity = getattr(self, name, None)
                if embedded_entity:
                    properties += prop.get_unindexed_properties(embedded_entity)

            elif not prop.indexed or prop.indexed_if and not prop.indexed_if(self, prop, name):
                properties += (prop.name_on_entity,)

        return properties

    @classmethod
    def pre_get_hook(cls, key):
        """A hook that runs before an entity is loaded from Datastore.
        Raising an exception here will prevent the entity from being
        loaded.

        Parameters:
          key(anom.Key): The datastore Key of the entity being loaded.
        """

    def post_get_hook(self):
        """A hook that runs after an entity has been loaded from
        Datastore.
        """

    @classmethod
    def get(cls, id_or_name, *, parent=None, namespace=None):
        """Get an entity by id.

        Parameters:
          id_or_name(int or str): The entity's id.
          parent(anom.Key, optional): The entity's parent Key.
          namespace(str, optional): The entity's namespace.

        Returns:
          Model: An entity or ``None`` if the entity doesn't exist in
          Datastore.
        """
        return Key(cls, id_or_name, parent=parent, namespace=namespace).get()

    @classmethod
    def pre_delete_hook(cls, key):
        """A hook that runs before an entity is deleted.  Raising an
        exception here will prevent the entity from being deleted.

        Parameters:
          key(anom.Key): The datastore Key of the entity being deleted.
        """

    @classmethod
    def post_delete_hook(cls, key):
        """A hook that runs after an entity has been deleted.

        Parameters:
          key(anom.Key): The datastore Key of the entity being deleted.
        """

    def delete(self):
        """Delete this entity from Datastore.

        Raises:
          RuntimeError: If this entity was never stored (i.e. if its
            key is partial).
        """
        return delete_multi([self.key])

    def pre_put_hook(self):
        """A hook that runs before this entity is persisted.  Raising
        an exception here will prevent the entity from being persisted.
        """

    def post_put_hook(self):
        """A hook that runs after this entity has been persisted.
        """

    def put(self):
        """Persist this entity to Datastore.
        """
        return put_multi([self])[0]

    @classmethod
    def query(cls, **options):
        """Return a new query for this Model.

        Parameters:
          \**options(dict): Parameters to pass to the :class:`Query`
            constructor.

        Returns:
          Query: The new query.
        """
        return Query(cls, **options)

    def __repr__(self):
        constructor = type(self).__name__
        properties = ("key",) + tuple(self._properties.keys())
        props = ", ".join(f"{name}={getattr(self, name)!r}" for name in properties)
        return f"{constructor}({props})"

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        if self is other:
            return True

        if self.key != other.key:
            return False

        for name in self._properties:
            if getattr(self, name) != getattr(other, name):
                return False

        return True

    def __ne__(self, other):
        return not (self == other)


def lookup_model_by_kind(kind):
    """Look up the model instance for a given Datastore kind.

    Parameters:
      kind(str)

    Raises:
      RuntimeError: If a model for the given kind has not been
        defined.

    Returns:
      model: The model class.
    """
    model = _known_models.get(kind)
    if model is None:
        raise RuntimeError(f"Model for kind {kind!r} not found.")
    return model


def delete_multi(keys):
    """Delete a set of entitites from Datastore by their
    respective keys.

    Note:
      This uses the adapter that is tied to the first model in the list.
      If the keys have disparate adapters this function may behave in
      unexpected ways.

    Warning:
      You must pass a **list** and not a generator or some other kind
      of iterable to this function as it has to iterate over the list
      of keys multiple times.

    Parameters:
      keys(list[anom.Key]): The list of keys whose entities to delete.

    Raises:
      RuntimeError: If the given set of keys have models that use
        a disparate set of adapters or if any of the keys are
        partial.
    """
    if not keys:
        return

    adapter = None
    for key in keys:
        if key.is_partial:
            raise RuntimeError(f"Key {key!r} is partial.")

        model = lookup_model_by_kind(key.kind)
        if adapter is None:
            adapter = model._adapter

        model.pre_delete_hook(key)

    adapter.delete_multi(keys)
    for key in keys:
        # Micro-optimization to avoid calling get_model.  This is OK
        # to do here because we've already proved that a model for
        # that kind exists in the previous block.
        model = _known_models[key.kind]
        model.post_delete_hook(key)


def get_multi(keys):
    """Get a set of entities from Datastore by their respective keys.

    Note:
      This uses the adapter that is tied to the first model in the
      list.  If the keys have disparate adapters this function may
      behave in unexpected ways.

    Warning:
      You must pass a **list** and not a generator or some other kind
      of iterable to this function as it has to iterate over the list
      of keys multiple times.

    Parameters:
      keys(list[anom.Key]): The list of keys whose entities to get.

    Raises:
      RuntimeError: If the given set of keys have models that use
        a disparate set of adapters or if any of the keys are
        partial.

    Returns:
      list[Model]: Entities that do not exist are going to be None
      in the result list.  The order of results matches the order
      of the input keys.
    """
    if not keys:
        return []

    adapter = None
    for key in keys:
        if key.is_partial:
            raise RuntimeError(f"Key {key!r} is partial.")

        model = lookup_model_by_kind(key.kind)
        if adapter is None:
            adapter = model._adapter

        model.pre_get_hook(key)

    entities_data, entities = adapter.get_multi(keys), []
    for key, entity_data in zip(keys, entities_data):
        if entity_data is None:
            entities.append(None)
            continue

        # Micro-optimization to avoid calling get_model.  This is OK
        # to do here because we've already proved that a model for
        # that kind exists in the previous block.
        model = _known_models[key.kind]
        entity = model._load(key, entity_data)
        entities.append(entity)
        entity.post_get_hook()

    return entities


def put_multi(entities):
    """Persist a set of entities to Datastore.

    Note:
      This uses the adapter that is tied to the first Entity in the
      list.  If the entities have disparate adapters this function may
      behave in unexpected ways.

    Warning:
      You must pass a **list** and not a generator or some other kind
      of iterable to this function as it has to iterate over the list
      of entities multiple times.

    Parameters:
      entities(list[Model]): The list of entities to persist.

    Raises:
      RuntimeError: If the given set of models use a disparate set of
        adapters.

    Returns:
      list[Model]: The list of persisted entitites.
    """
    if not entities:
        return []

    adapter, requests = None, []
    for entity in entities:
        if adapter is None:
            adapter = entity._adapter

        entity.pre_put_hook()
        requests.append(PutRequest(entity.key, entity.unindexed_properties, entity))

    keys = adapter.put_multi(requests)
    for key, entity in zip(keys, entities):
        entity.key = key
        entity.post_put_hook()

    return entities



import warnings

from .datastore_adapter import DatastoreAdapter  # noqa

try:
    from .memcache_adapter import MemcacheAdapter  # noqa
except ImportError:  # pragma: no cover
    warnings.warn(
        "MemcacheAdapter could not be imported.  Install anom with `pip "
        "install anom[memcache]` to add memcache support.",
        ImportWarning, stacklevel=2
    )



import base64
import json
import msgpack
import operator
import zlib

from collections import defaultdict
from copy import copy
from datetime import datetime
from dateutil import tz
from enum import IntEnum
from functools import reduce
from itertools import chain

from . import model
from .model import EmbedLike, Property, NotFound, Skip, classname


#: The UNIX epoch.
_epoch = datetime(1970, 1, 1, tzinfo=tz.tzutc())


def _seconds_since_epoch(dt):
    return (dt - _epoch).total_seconds()


#: The maximum length of indexed properties.
_max_indexed_length = 1500


class Blob:
    """Mixin for Properties whose values cannot be indexed.
    """

    def __init__(self, **options):
        if options.get("indexed") or options.get("indexed_if"):
            raise TypeError(f"{classname(self)} properties cannot be indexed.")

        super().__init__(**options)


class Compressable(Blob):
    """Mixin for Properties whose values can be gzipped before being
    persisted.

    Parameters:
      compressed(bool): Whether or not values belonging to this
        Property should be stored gzipped in Datastore.
      compression_level(int): The amount of compression to apply.
        See :func:`zlib.compress` for details.
    """

    def __init__(self, *, compressed=False, compression_level=-1, **options):
        if not (-1 <= compression_level <= 9):
            raise ValueError("compression_level must be an integer between -1 and 9.")

        super().__init__(**options)

        self.compressed = compressed
        self.compression_level = compression_level

    def prepare_to_load(self, entity, value):
        if value is not None and self.compressed:
            value = zlib.decompress(value)

        return super().prepare_to_load(entity, value)

    def prepare_to_store(self, entity, value):
        if value is not None and self.compressed:
            value = zlib.compress(value, level=self.compression_level)

        return super().prepare_to_store(entity, value)


class Encodable:
    """Mixin for string properties that have an encoding.

    Parameters:
      encoding(str): The encoding to use when persisting this Property
        to Datastore.  Defaults to ``utf-8``.
    """

    def __init__(self, *, encoding="utf-8", **options):
        super().__init__(**options)

        self.encoding = encoding

    def prepare_to_load(self, entity, value):
        value = super().prepare_to_load(entity, value)

        # BUG(gcloud): Projections seem to cause bytes to be
        # loaded as strings so this instance check is required.
        if value is not None and isinstance(value, (list, bytes)):
            if self.repeated:
                value = [v.decode(self.encoding) for v in value]
            else:
                value = value.decode(self.encoding)

        return value

    def prepare_to_store(self, entity, value):
        if value is not None:
            if self.repeated:
                value = [v.encode(self.encoding) for v in value]
            else:
                value = value.encode(self.encoding)

        return super().prepare_to_store(entity, value)


class Serializer(Compressable, Property):
    """Base class for properties that serialize data.
    """

    _types = (bool, bytes, dict, float, int, str, datetime, model.Key, model.Model)

    @classmethod
    def _serialize(cls, value):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _deserialize(cls, data):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _dumps(cls, value):  # pragma: no cover
        raise NotImplementedError

    @classmethod
    def _loads(cls, data):  # pragma: no cover
        raise NotImplementedError

    @staticmethod
    def _entity_to_dict(entity):
        return {
            "key": entity.key,
            "kind": entity._kind,
            "data": dict(entity),
        }

    @staticmethod
    def _entity_from_dict(data):
        key = model.Key(*data["key"])
        clazz = model.lookup_model_by_kind(data["kind"])
        instance = clazz._load(key, data["data"])
        return instance

    def prepare_to_load(self, entity, value):
        if value is not None:
            value = self._loads(value)

        return super().prepare_to_load(entity, value)

    def prepare_to_store(self, entity, value):
        if value is not None:
            value = self._dumps(value)

        return super().prepare_to_store(entity, value)


class Bool(Property):
    """A Property for boolean values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    _types = (bool,)

    @property
    def is_true(self):
        "PropertyFilter: A filter that checks if this value is True."
        return self == True  # noqa

    @property
    def is_false(self):
        "PropertyFilter: A filter that checks if this value is False."
        return self == False  # noqa


class Bytes(Compressable, Property):
    """A Property for bytestring values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
      compressed(bool, optional): Whether or not this property should
        be compressed before being persisted.
      compression_level(int, optional): The amount of compression to
        apply when compressing values.
    """

    _types = (bytes,)


class Computed(Property):
    """A Property for values that should be computed dinamically based
    on the state of the entity.  Values on an entity are only computed
    the first time computed properties are accessed on that entity and
    they are re-computed every time the entity is loaded from
    Datastore.

    Computed properties cannot be assigned to and their "cache" can be
    busted by deleting them::

      del an_entity.a_property

    Note:

      Computed properties are **indexed** and **optional** by default
      for convenience.  This is different from all other built-in
      properties.

    Parameters:
      fn(callable): The function to use when computing the data.
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``True``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``True``.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.
    """

    _types = (object,)

    def __init__(self, fn, **options):
        # Computed properties are/should mainly be used for filtering
        # purposes, so it makes sense for them to default to being
        # both optional and indexed for convenience.
        options.setdefault("indexed", True)
        options.setdefault("optional", True)

        super().__init__(**options)

        self.fn = fn

    def __get__(self, ob, obtype):
        if ob is None:
            return self

        value = ob._data.get(self.name_on_model, NotFound)
        if value is NotFound:
            value = ob._data[self.name_on_model] = self.fn(ob)

        return value

    def __set__(self, ob, value):
        raise AttributeError("Can't set attribute.")

    def prepare_to_load(self, entity, value):
        return Skip


class DateTime(Property):
    """A Property for :class:`datetime.datetime` values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
      auto_now_add(bool, optional): Whether or not to set this
        property's value to the current time the first time it's
        stored.
      auto_now(bool, optional): Whether or not this property's value
        should be set to the current time every time it is stored.
    """

    _types = (datetime,)

    def __init__(self, *, auto_now_add=False, auto_now=False, **options):
        super().__init__(**options)

        self.auto_now_add = auto_now_add
        self.auto_now = auto_now

        if self.repeated and (auto_now_add or auto_now):
            raise TypeError("Cannot use auto_now{,_add} with repeated properties.")

    def _current_value(self):
        return datetime.now(tz.tzlocal())

    def prepare_to_load(self, entity, value):
        return super().prepare_to_load(entity, value)

    def prepare_to_store(self, entity, value):
        if value is None and self.auto_now_add:
            value = entity._data[self.name_on_model] = self._current_value()
        elif self.auto_now:
            value = entity._data[self.name_on_model] = self._current_value()

        if value is not None:
            value = entity._data[self.name_on_model] = value.astimezone(tz.tzutc())

        return super().prepare_to_store(entity, value)

    def validate(self, value):
        value = super().validate(value)
        if value is not None and not value.tzinfo:
            return value.replace(tzinfo=tz.tzlocal())
        return value


class Float(Property):
    """A Property for floating point values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    _types = (float,)


class Integer(Property):
    """A Property for integer values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    _types = (int,)


class Json(Serializer):
    """A Property for values that should be stored as JSON.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      compressed(bool, optional): Whether or not this property should
        be compressed before being persisted.
      compression_level(int, optional): The amount of compression to
        apply when compressing values.
    """

    #: The name of the field that is used to store type information
    #: about non-standard JSON values.
    _type_field = "__anom_type"

    @classmethod
    def _serialize(cls, value):
        if isinstance(value, bytes):
            kind, value = "blob", base64.b64encode(value).decode("utf-8")

        elif isinstance(value, datetime):
            kind, value = "datetime", _seconds_since_epoch(value)

        elif isinstance(value, model.Model):
            kind, value = "model", cls._entity_to_dict(value)

        else:
            raise TypeError(f"Value of type {type(value)} cannot be serialized.")

        return {cls._type_field: kind, "value": value}

    @classmethod
    def _deserialize(cls, data):
        kind = data.get(cls._type_field, None)
        if kind is None:
            return data

        value = data.get("value")
        if kind == "blob":
            return base64.b64decode(value.encode("utf-8"))

        elif kind == "datetime":
            return datetime.fromtimestamp(value, tz.tzutc())

        elif kind == "model":
            return cls._entity_from_dict(value)

        else:
            raise ValueError(f"Invalid kind {kind!r}.")

    @classmethod
    def _dumps(cls, value):
        return json.dumps(value, separators=(",", ":"), default=Json._serialize)

    @classmethod
    def _loads(cls, data):
        return json.loads(data, object_hook=Json._deserialize)


class Key(Property):
    """A Property for :class:`anom.Key` values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      kind(str or model, optional): The kinds of keys that may be
        assigned to this property.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    _types = (model.Model, model.Key, tuple)

    def __init__(self, *, kind=None, **options):
        super().__init__(**options)

        if isinstance(kind, model.model):
            self.kind = kind._kind
        else:
            self.kind = kind

    def validate(self, value):
        value = super().validate(value)

        if value is not None:
            if self.repeated and isinstance(value, list):
                return [self.validate(v) for v in value]

            if isinstance(value, model.Model):
                value = value.key

            # Serialized keys may end up being turned into tuples
            # (JSON, msgpack) so we attempt to coerce them here.
            if isinstance(value, tuple):
                value = model.Key(*value)

            if value.is_partial:
                raise ValueError("Cannot assign partial Keys to Key properties.")

            elif self.kind and self.kind != value.kind:
                raise ValueError(f"Property {self.name_on_model} cannot be assigned keys of kind {value.kind}.")

        return value


class Msgpack(Serializer):
    """A Property for values that should be stored as msgpack data.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      compressed(bool, optional): Whether or not this property should
        be compressed before being persisted.
      compression_level(int, optional): The amount of compression to
        apply when compressing values.
    """

    class Extensions(IntEnum):
        Model = 0
        DateTime = 1

    @classmethod
    def _serialize(cls, value):
        if isinstance(value, model.Model):
            kind, value = cls.Extensions.Model, cls._entity_to_dict(value)

        elif isinstance(value, datetime):
            kind, value = cls.Extensions.DateTime, _seconds_since_epoch(value)

        else:
            raise TypeError(f"Value of type {type(value)} cannot be serialized.")

        return msgpack.ExtType(kind, cls._dumps(value))

    @classmethod
    def _deserialize(cls, code, data):
        try:
            kind = cls.Extensions(code)
        except ValueError:
            raise ValueError(f"Invalid extension code {code}.")

        value = cls._loads(data)
        if kind is cls.Extensions.Model:
            return cls._entity_from_dict(value)

        elif kind is cls.Extensions.DateTime:
            return datetime.fromtimestamp(value, tz.tzutc())

    @classmethod
    def _dumps(cls, value):
        return msgpack.packb(value, default=cls._serialize, use_bin_type=True)

    @classmethod
    def _loads(cls, data):
        return msgpack.unpackb(data, ext_hook=cls._deserialize, encoding="utf-8")


class String(Encodable, Property):
    """A Property for indexable string values.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
      encoding(str): The encoding to use when persisting this Property
        to Datastore.  Defaults to ``utf-8``.
    """

    _types = (str,)

    def _validate_length(self, value):
        if len(value) > _max_indexed_length and \
           len(value.encode(self.encoding)) > _max_indexed_length:
            raise ValueError(
                f"String value is longer than the maximum allowed length "
                f"({_max_indexed_length}) for indexed properties. Set "
                f"indexed to False if the value should not be indexed."
            )

    def validate(self, value):
        value = super().validate(value)
        if not self.indexed or value is None:
            return value

        if not self.repeated:
            self._validate_length(value)

        return value


class Text(Encodable, Compressable, Property):
    """A Property for long string values that are never indexed.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
      compressed(bool, optional): Whether or not this property should
        be compressed before being persisted.
      compression_level(int, optional): The amount of compression to
        apply when compressing values.
      encoding(str): The encoding to use when persisting this Property
        to Datastore.  Defaults to ``utf-8``.
    """

    _types = (str,)


class Unicode(Property):
    """A Property for string values that should be stored as unicode strings.

    Parameters:
      name(str, optional): The name of this property on the Datastore
        entity.  Defaults to the name of this property on the model.
      default(object, optional): The property's default value.
      indexed(bool, optional): Whether or not this property should be
        indexed.  Defaults to ``False``.
      indexed_if(callable, optional): Whether or not this property
        should be indexed when the callable returns ``True``.
        Defaults to ``None``.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    _types = (str,)

    def _validate_length(self, value):
        if len(value.encode('utf-8')) > _max_indexed_length:
            raise ValueError(
                f"Unicode value is longer than the maximum allowed length "
                f"({_max_indexed_length} bytes) for indexed properties. Set "
                f"indexed to False if the value should not be indexed."
            )

    def validate(self, value):
        value = super().validate(value)
        if not self.indexed or value is None:
            return value

        if not self.repeated:
            self._validate_length(value)

        return value


class Embed(EmbedLike):
    """A property for embedding entities inside other entities.

    Parameters:
      kind(str or model): The kinds of keys that may be assigned to
        this property.
      optional(bool, optional): Whether or not this property is
        optional.  Defaults to ``False``.  Required but empty values
        cause models to raise an exception before data is persisted.
      repeated(bool, optional): Whether or not this property is
        repeated.  Defaults to ``False``.  Optional repeated
        properties default to an empty list.
    """

    def __init__(self, *, kind, **options):
        if "default" in options or "indexed" in options or "indexed_if" in options:
            raise TypeError(f"{classname(self)} does not support name, default, indexed or indexed_if.")

        super().__init__(**options)

        if isinstance(kind, model.model):
            self.kind = kind._kind
        else:
            self.kind = kind

    def __copy__(self):
        prop = type(self)(kind=self.kind)
        for name in vars(self):
            setattr(prop, name, getattr(self, name))

        return prop

    def get_unindexed_properties(self, entity):
        if isinstance(entity, list):
            return tuple(set(chain.from_iterable(self.get_unindexed_properties(e) for e in entity)))
        return tuple(f"{self.name_on_entity}.{name}" for name in entity.unindexed_properties)

    def validate(self, value):
        if self.optional and value is None:
            return [] if self.repeated else None

        model_class = model.lookup_model_by_kind(self.kind)
        if self.repeated and not all(isinstance(v, model_class) for v in value):
            raise TypeError(f"{self.name_on_model} properties must be instances of {self.kind}.")

        elif not self.repeated and not isinstance(value, model_class):
            raise TypeError(f"{self.name_on_model} properties must be instances of {self.kind}.")

        return value

    def prepare_to_load(self, entity, data):
        embed_prefix = f"{self.name_on_entity}."
        embed_prefix_len = len(embed_prefix)
        data = {k[embed_prefix_len:]: v for k, v in data.items() if k.startswith(embed_prefix)}
        if self.repeated:
            return self._prepare_to_load_repeated_properties(data)
        return self._prepare_to_load_properties(data)

    def _prepare_to_load_repeated_properties(self, data):
        if self.optional and not data:
            return []

        datas = [{} for _ in range(len(next(iter(data.values()))))]
        for key, values in data.items():
            for i, value in enumerate(values):
                datas[i][key] = value

        return [self._prepare_to_load_properties(data) for data in datas]

    def _prepare_to_load_properties(self, data):
        if self.optional and not data:
            return None

        model_class = model.lookup_model_by_kind(self.kind)
        model_key = model.Key(self.kind)
        return model_class._load(model_key, data)

    def prepare_to_store(self, entity, value):
        if value is not None:
            if isinstance(value, list):
                yield from self._prepare_to_store_repeated_properties(value)

            else:
                yield from self._prepare_to_store_properties(value)

        elif not self.optional:
            raise RuntimeError(f"Property {self.name_on_model} requires a value.")

    def _prepare_to_store_repeated_properties(self, entities):
        if self.optional and not entities:
            return []

        properties = defaultdict(list)
        for entity in entities:
            for name, value in self._prepare_to_store_properties(entity):
                properties[name].append(value)

        # Ensure all sublists are equal, otherwise rebuilding the
        # entity is not going to be possible.
        props_are_valid = reduce(operator.eq, (len(val) for val in properties.values()))
        if not props_are_valid:  # pragma: no cover
            raise ValueError(
                "Repeated properties for {self.name_on_model} have different lengths. "
                "You probably just found a bug in anom, please report it!"
            )

        for name, values in properties.items():
            yield name, values

    def _prepare_to_store_properties(self, entity):
        for name, prop in entity._properties.items():
            value = getattr(entity, name)
            if isinstance(prop, EmbedLike):
                for name, value in prop.prepare_to_store(entity, value):
                    yield f"{self.name_on_entity}.{name}", value
            else:
                yield f"{self.name_on_entity}.{prop.name_on_entity}", prop.prepare_to_store(entity, value)

    def __getattr__(self, name):
        model_class = model.lookup_model_by_kind(self.kind)
        value = getattr(model_class, name)
        if isinstance(value, model.Property):
            value = copy(value)
            value._name_on_entity = f"{self.name_on_entity}.{value.name_on_entity}"
            value._name_on_model = f"{self.name_on_model}.{value.name_on_model}"
            return value
        return value



from setuptools import setup


with open("anom/__init__.py") as module:
    version_marker = "__version__ = "
    for line in module:
        if line.startswith(version_marker):
            version = line[len(version_marker):].strip().strip('"')
            break
    else:
        assert False, "could not find version in anom/__init__.py"


dependencies = []
with open("requirements.txt") as reqs:
    for line in reqs:
        dependencies.append(line.strip())


extra_dependencies = {}
for group in ("memcache",):
    extra_dependencies[group] = extra_dep_list = []
    with open(f"requirements-{group}.txt") as reqs:
        for line in reqs:
            extra_dep_list.append(line.strip())


setup(
    name="anom",
    version=version,
    description="anom is an object mapper for Google Cloud Datastore.",
    long_description="https://github.com/Bogdanp/anom-py",
    packages=["anom", "anom.adapters", "anom.testing"],
    install_requires=dependencies,
    extras_require=extra_dependencies,
    author="Bogdan Popa",
    author_email="popa.bogdanp@gmail.com",
    url="https://github.com/Bogdanp/anom-py",
)



def is_default(entity, prop, name):
    "bool: True if the value of a property is equal to its default."
    return getattr(entity, name) == prop.default


def is_not_default(entity, prop, name):
    "bool: True if the value of a property is not equal to its default."
    return getattr(entity, name) != prop.default


def is_empty(entity, prop, name):
    "bool: True if the value of a property is not set."
    return name not in entity._data


def is_not_empty(entity, prop, name):
    "bool: True if the value of a  property is set."
    return name in entity._data


def is_none(entity, prop, name):
    "bool: True if the value of a property is None."
    return is_not_empty(entity, prop, name) and getattr(entity, name) is None


def is_not_none(entity, prop, name):
    "bool: True if the value of a property is not None."
    return is_not_empty(entity, prop, name) and name in entity._data and getattr(entity, name) is not None


def is_true(entity, prop, name):
    "bool: True if the value of a property is True."
    return is_not_empty(entity, prop, name) and name in entity._data and bool(getattr(entity, name))


def is_false(entity, prop, name):
    "bool: True if the value of a property is False."
    return is_not_empty(entity, prop, name) and name in entity._data and not bool(getattr(entity, name))



from collections import namedtuple

from .namespaces import get_namespace


DEFAULT_BATCH_SIZE = 300


class PropertyFilter(namedtuple("PropertyFilter", ("name", "operator", "value"))):
    """Represents an individual filter on a Property within a Query.
    """


class QueryOptions(dict):
    """Options that determine how data is fetched from the Datastore
    on a per Query basis.

    Parameters:
      batch_size(int, optional): The number of results to fetch per batch.
      keys_only(bool, optional): Whether or not the results should be
        Keys or Entities.
      limit(int, optional): The maximum number of results to return.
      offset(int, optional): The number of results to skip.
      cursor(str, optional): A url-safe cursor representing where in
        the result set the query should start.
    """

    def __init__(self, query, **options):
        super().__init__(**options)

        self.query = query

    def replace(self, **options):
        """Update this options object in place.

        Parameters:
          \**options(QueryOptions)

        Returns:
          QueryOptions: The updated instance.
        """
        self.update(options)
        return self

    @property
    def batch_size(self):
        """int: The number of results to fetch per batch.  Clamped to
        limit if limit is set and is smaller than the given batch
        size.
        """
        batch_size = self.get("batch_size", DEFAULT_BATCH_SIZE)
        if self.limit is not None:
            return min(self.limit, batch_size)
        return batch_size

    @property
    def keys_only(self):
        "bool: Whether or not the results should be Keys or Entities."
        return self.get("keys_only", False)

    @property
    def limit(self):
        "int: The maximum number of results to return."
        return self.get("limit", self.query.limit)

    @property
    def offset(self):
        "int: The number of results to skip."
        return self.get("offset", self.query.offset)

    @property
    def cursor(self):
        "bytes: The url-safe cursor for a query."
        return self.get("cursor", b"")

    @cursor.setter
    def cursor(self, value):
        self["cursor"] = value


class Resultset:
    """An iterator for datastore query results.

    Parameters:
     query(Query): The query that was run to create this resultset.
     options(QueryOptions): Options that determine how entities are
       fetched from Datastore.
    """

    def __init__(self, query, options):
        self._query = query
        self._options = options
        self._complete = False
        self._entities = self._get_entities()

    @property
    def cursor(self):
        "str: The url-safe cursor for the next batch of results."
        return self._options.cursor

    @property
    def has_more(self):
        "bool: Whether or not there are more results."
        return not self._complete

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._entities)

    def _get_batches(self):
        from .adapter import get_adapter

        remaining = self._options.limit
        while True:
            adapter = self._query.model._adapter if self._query.model else get_adapter()
            entities, self._options.cursor = adapter.query(self._query, self._options)
            if remaining is not None:
                remaining -= len(entities)
                if remaining < 0:
                    entities = entities[:remaining]

            # This check is here for compatibility with the datastor emulator.
            if not entities:
                break

            # If we received fewer entities than we asked for then we
            # can safely say that we've finished iterating.  We have
            # to do this before yielding, however.
            if len(entities) < self._options.batch_size:
                self._complete = True

            if self._options.keys_only:
                yield (key for key, _ in entities)
            else:
                yield (key.get_model()._load(key, data) for key, data in entities)

            # Datastore now returns None as the next cursor if there
            # are no more values.  The emulator, though, behaves the
            # same way it used to behave so we need both this check as
            # well as the "if not entities" check above.
            if self._options.cursor is None or remaining is not None and remaining <= 0:
                break

        self._complete = True

    def _get_entities(self):
        for batch in self._get_batches():
            yield from batch


class Page:
    """An iterator that represents a single page of entities or keys.

    Parameters:
      cursor(str): The url-safe cursor for the next page of results.
      batch(iterator[Model or anom.Key]): The batch of results backing
        this Page.
    """

    def __init__(self, cursor, batch):
        self._cursor = cursor
        self._batch = batch

    @property
    def cursor(self):
        "str: The url-safe cursor for the next page of results."
        return self._cursor

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._batch)


class Pages:
    """An iterator for :class:`Pages<Page>` of results.

    Parameters:
     query(Query): The query that was run to create this resultset.
     options(QueryOptions): Options that determine how entities are
       fetched from Datastore.
    """

    def __init__(self, query, page_size, options):
        options = QueryOptions(query, **options)
        options.update(batch_size=page_size)

        self._resultset = Resultset(query, options)
        self._pages = self._get_pages()

    @property
    def has_more(self):
        "bool: Whether or not there are more pages."
        return self._resultset.has_more

    @property
    def cursor(self):
        "str: The url-safe cursor for the next page of results."
        return self._resultset.cursor

    def fetch_next_page(self):
        """Fetch the next Page of results.

        Returns:
          Page: The next page of results.
        """
        for page in self:
            return page
        else:
            return Page(self._resultset.cursor, iter(()))

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._pages)

    def _get_pages(self):
        for batch in self._resultset._get_batches():
            yield Page(self._resultset.cursor, batch)


class Query(namedtuple("Query", (
    "model", "kind", "ancestor", "namespace", "projection", "filters", "orders", "offset", "limit",
))):
    """An immutable Datastore query.

    Parameters:
      kind(str or model): The Datastore kind to query.
      ancestor(anom.Key, optional): The ancestor to which this query should be scoped.
      namespace(str, optional): The namespace to which this query should be scoped.
      projection(tuple[str], optional): The tuple or tuple of fields to project.
      filters(tuple[PropertyFilter], optional): The tuple of filters to apply.
      orders(tuple[str], optional): The tuple of sort orders to apply.
      offset(int, optional): The number of query results to skip.
      limit(int, optional): The maximum number of results to return.

    Example:

      You can construct queries declaratively::

        people_query = Query(Person)
          .where(Person.email == "test@example.com")
          .and_where(Person.enabled.is_true)
          .with_limit(10)

      Then run them to iterate over all the results::

        all_people = people_query.run()
        for person in all_people:
          print(person)

      Or paginate over them::

        for page in people_query.paginate(page_size=10):
          print("Cursor: ", page.cursor)
          for person in page:
            print(person)

      Or get individual pages of results::

        page_1 = people_query.paginate(page_size=10).fetch_next_page()
        page_2 = people_query.paginate(page_size=10, cursor=page_1.cursor).fetch_next_page()

    """

    def __new__(
            cls, kind=None, *, ancestor=None, namespace=None,
            projection=(), filters=(), orders=(), offset=0, limit=None,
    ):
        from .model import lookup_model_by_kind

        if kind is None:
            model = None

        elif isinstance(kind, str):
            model = lookup_model_by_kind(kind)

        else:
            model, kind = kind, kind._kind

        if namespace is None:
            namespace = get_namespace()

        return super().__new__(
            cls, model=model, kind=kind, ancestor=ancestor, namespace=namespace,
            projection=_prepare_projection(projection), filters=tuple(filters), orders=tuple(orders),
            offset=offset, limit=limit,
        )

    def select(self, *projection):
        """Return a new query with its projection replaced.

        Parameters:
          \*projection(str): The fields to project.

        Returns:
          Query: The derived Query.
        """
        return self._replace(projection=_prepare_projection(projection))

    def where(self, *filters):
        """Return a new query, replacing the current set of filters.

        Parameters:
          \*filters(PropertyFilter): The filters to add.

        Returns:
          Query: The derived Query.
        """
        return self._replace(filters=filters)

    def and_where(self, *filters):
        """Return a new query, adding the given filters with the
        current query's filters to form an "and".

        Parameters:
          \*filters(PropertyFilter): The filters to add.

        Returns:
          Query: The derived Query.
        """
        return self._replace(filters=self.filters + filters)

    def order_by(self, *orders):
        """Returns a new query containing an additional set of orders.

        Parameters:
          \*orders(str): The sort orders to add.

        Returns:
          Query: The derived Query.
        """
        return self._replace(orders=self.orders + orders)

    def with_ancestor(self, ancestor):
        """Returns a new query with its ancestor updated.

        Parameters:
          ancestor(anom.Key): The new ancestor.

        Returns:
          Query: The derived Query.
        """
        return self._replace(ancestor=ancestor)

    def with_namespace(self, namespace):
        """Returns a new query with its namespace updated.

        Parameters:
          namespace(str): The new namespace.

        Returns:
          Query: The derived Query.
        """
        return self._replace(namespace=namespace)

    def with_offset(self, offset):
        """Returns a new query with its offset updated.

        Parameters:
          offset(int): The new offset.

        Returns:
          Query: The derived Query.
        """
        return self._replace(offset=offset)

    def with_limit(self, limit):
        """Returns a new query with its limit updated.

        Parameters:
          limit(int): The new limit.

        Returns:
          Query: The derived Query.
        """
        return self._replace(limit=limit)

    def count(self, *, page_size=DEFAULT_BATCH_SIZE, **options):
        """Counts the number of entities that match this query.

        Note:
          Since Datastore doesn't provide a native way to count
          entities by query, this method paginates through all the
          entities' keys and counts them.

        Parameters:
          \**options(QueryOptions, optional)

        Returns:
          int: The number of entities.
        """
        entities = 0
        options = QueryOptions(self).replace(keys_only=True)
        for page in self.paginate(page_size=page_size, **options):
            entities += len(list(page))
        return entities

    def delete(self, *, page_size=DEFAULT_BATCH_SIZE, **options):
        """Deletes all the entities that match this query.

        Note:
          Since Datasotre doesn't provide a native way to delete
          entities by query, this method paginates through all the
          entities' keys and issues a single delete_multi call per
          page.

        Parameters:
          \**options(QueryOptions, optional)

        Returns:
          int: The number of deleted entities.
        """
        from .model import delete_multi

        deleted = 0
        options = QueryOptions(self).replace(keys_only=True)
        for page in self.paginate(page_size=page_size, **options):
            keys = list(page)
            deleted += len(keys)
            delete_multi(keys)

        return deleted

    def get(self, **options):
        """Run this query and get the first result.

        Parameters:
          \**options(QueryOptions, optional)

        Returns:
          Model: An entity or None if there were no results.
        """
        sub_query = self.with_limit(1)
        options = QueryOptions(sub_query).replace(batch_size=1)
        for result in sub_query.run(**options):
            return result
        return None

    def run(self, **options):
        """Run this query and return a result iterator.

        Parameters:
          \**options(QueryOptions, optional)

        Returns:
          Resultset: An iterator for this query's results.
        """
        return Resultset(self._prepare(), QueryOptions(self, **options))

    def paginate(self, *, page_size, **options):
        """Run this query and return a page iterator.

        Parameters:
          page_size(int): The number of entities to fetch per page.
          \**options(QueryOptions, optional)

        Returns:
          Pages: An iterator for this query's pages of results.
        """
        return Pages(self._prepare(), page_size, QueryOptions(self, **options))

    def _prepare(self):
        # Polymorphic children need to be able to query for themselves
        # and their subclasses.
        if self.model and self.model._is_child:
            kind_filter = (self.model._kinds_name, "=", self.model._kinds[0])
            return self._replace(filters=(kind_filter,) + self.filters)

        return self


def _prepare_projection(projection):
    return tuple(f if isinstance(f, str) else f.name_on_entity for f in projection)
